// ==============================================================
//                 ORBITER MODULE: MARS PHOENIX 
//                  
//
// mpx.cpp
// Control module for MARS PHOENIX vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double PA_ISP = 2500;
const double PB_MAXRCSTH = 4.5;
const double PB_MAXRCSTH20 = 90;
const double PA_FUELMASS = 62;
const double MPX_BKSHLL_MASS = 80;
const double MPX_SOLP_MASS = 100;
const double MPX_HTSHLD_MASS = 40;
const double MPX_LNDR_MASS = 425;
const double ANTENNA_OPERATING_SPEED = 0.25;
const double ARRAY_OPERATING_SPEED = 0.5;
const double LEGS_OPERATING_SPEED = 1;
const double COL_OPERATING_SPEED = 0.25;


//DWORD texidx = 15;
static const DWORD ntdvtx_geardown = 4;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(   0, 2,0.1), 1e5, 1e4, 0.5, 0.5},
	{_V(-1.732,-1,0.1), 1e5, 1e4, 0.5, 0.5},
	{_V( 1.732,-1,0.1), 1e5, 1e4, 0.5, 0.5},
	{_V(   0,  0,  -1), 1e5, 1e4, 0.5}
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}


class mpx: public VESSEL3 {
public:
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } ant_status, fold_status, legs_status, col_status;
	mpx (OBJHANDLE hVessel, int flightmodel);
		~mpx ();
	
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);

	private:

	void DefineAnimations (void);
	void RevertAntenna (void);
	void RevertArray (void);
	void RevertLegs (void);
	void RevertCol (void);
	void DoChutes(void);
	void DoCol(void);
	void ApplySkin();
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	double CalcEmptyMass();
	void SpawnObject(char* classname, char* ext, VECTOR3 ofs, VECTOR3 vofs);
	void SpawnObject2(char* classname, char* ext, VECTOR3 ofs2, VECTOR3 vofs2);
	//void SpawnObject3(char* classname, char* ext);
	void Jettison();
	void Jettison2();
	void Jettison3();

	THRUSTER_HANDLE th_rcs[14], th_group[2];
	PROPELLANT_HANDLE rcspr;
	MESHHANDLE hmpx_backshellmesh, hmpx_landermesh, hmpx_heatshieldmesh;
	ANIMATIONCOMPONENT_HANDLE Parent, Parent1, Parent2, Parent3, Parent4, Parent5;
	SURFHANDLE bs1, bs2, bs3, bs4;
	ATTACHMENTHANDLE hAttach, hDelta;

	UINT anim_ant, anim_fold, anim_legs, anim_col;
	double ant_proc, fold_proc, legs_proc, col_proc;

	//int rocksw, texsw, status;
	int texsw, status;
	bool is_attached;
};

mpx::mpx (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

mpx::~mpx ()
{

}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void mpx::DefineAnimations (void)
{
	// 1. solp
	static UINT HiGainAnt1Grp[2] = {0,8};
	static MGROUP_ROTATE HiGainAnt1 (2, HiGainAnt1Grp, 2, _V(-1.077,0,-1.241), _V(0,1,0), (float)(PI));
	static UINT HiGainAnt2Grp[2] = {5,11};
	static MGROUP_ROTATE HiGainAnt2 (2, HiGainAnt2Grp, 2, _V(1.077,0,-1.241), _V(0,-1,0), (float)(PI));
	anim_ant = CreateAnimation (0);
	AddAnimationComponent (anim_ant, 0.1, 1, &HiGainAnt1);
	AddAnimationComponent (anim_ant, 0.1, 1,   &HiGainAnt2);

	
	// 2. chutes
	anim_fold = CreateAnimation (0);
	static UINT ArrayLFoldGrp[1] = {0};
	static UINT ArrayRFoldGrp[1] = {5};
	
	static MGROUP_SCALE ArrayLFold1 (0, ArrayLFoldGrp, 1, _V(0,0,-0.8), _V(12.5,12.5,150));
	AddAnimationComponent (anim_fold, 0,   0.25, &ArrayLFold1);
	static MGROUP_SCALE ArrayLFold2 (0, ArrayLFoldGrp, 1, _V(0,0,-0.8), _V(12,12,1));
	AddAnimationComponent (anim_fold, 0.25,   0.5, &ArrayLFold2);
	static MGROUP_SCALE ArrayLFold3 (0, ArrayRFoldGrp, 1, _V(0,0,-1), _V(0.01,0.01,0.01));
	AddAnimationComponent (anim_fold, 0,   0.1, &ArrayLFold3);


	// 3. legs
	anim_legs = CreateAnimation (0);
	static UINT Legs1[3] = {1,28,34};
	static MGROUP_TRANSLATE LegsAnim1 (3, Legs1, 3, _V(0,0.144,0.24941376));
	AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1);

	static UINT Legs1LB[1] = {23};
	static MGROUP_ROTATE LegsAnim1lb (3, Legs1LB, 1, _V(-0.574,0.299,0.176), _V(0.7165,-0.6041,0.3488),(float)(PI/12.5));
	Parent = AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1lb);
	static UINT Legs1LG[1] = {22};
	static MGROUP_TRANSLATE LegsAnim1lg (3, Legs1LG, 1, _V(0.0756,0.0948,0.00888));
	AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1lg, Parent);

	static UINT Legs1RB[1] = {12};
	static MGROUP_ROTATE LegsAnim1rb (3, Legs1RB, 1, _V(0.574,0.299,0.176), _V(-0.7165,-0.6041,0.3488),(float)(-PI/12.5));
	Parent1 = AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1rb);
	static UINT Legs1RG[1] = {21};
	static MGROUP_TRANSLATE LegsAnim1rg (3, Legs1RG, 1, _V(-0.0756,0.0948,0.00888));
	AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1rg, Parent1);


	static UINT Legs2[3] = {2,29,35};
	static MGROUP_TRANSLATE LegsAnim2 (3, Legs2, 3, _V(0.12470688,-0.072,0.24941376));
	AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2);

	static UINT Legs2LB[1] = {14};
	static MGROUP_ROTATE LegsAnim2lb (3, Legs2LB, 1, _V(0.545,0.348,0.176), _V(-0.8863,-0.322,0.3329),(float)(PI/12.5));
	Parent2 = AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2lb);
	static UINT Legs2LG[1] = {18};
	static MGROUP_TRANSLATE LegsAnim2lg (3, Legs2LG, 1, _V(0.0444,-0.11304,0.00888));
	AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2lg, Parent2);

	static UINT Legs2RB[1] = {13};
	static MGROUP_ROTATE LegsAnim2rb (3, Legs2RB, 1, _V(-0.028,-0.646,0.176), _V(0.1639,-0.9287,-0.3328),(float)(PI/12.5));
	Parent3 = AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2rb);
	static UINT Legs2RG[1] = {17};
	static MGROUP_TRANSLATE LegsAnim2rg (3, Legs2RG, 1, _V(0.12,0.018,0.00888));
	AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2rg, Parent3);




	static UINT Legs3[3] = {3,30,33};
	static MGROUP_TRANSLATE LegsAnim3 (3, Legs3, 3, _V(-0.12470688,-0.072,0.24941376));
	AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3);


	static UINT Legs3LB[1] = {15};
	static MGROUP_ROTATE LegsAnim3lb (3, Legs3LB, 1, _V(-0.545,0.348,0.176), _V(0.8863,-0.322,0.3329),(float)(-PI/12.5));
	Parent4 = AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3lb);
	static UINT Legs3LG[1] = {19};
	static MGROUP_TRANSLATE LegsAnim3lg (3, Legs3LG, 1, _V(-0.0444,-0.11304,0.00888));
	AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3lg, Parent4);

	static UINT Legs3RB[1] = {16};
	static MGROUP_ROTATE LegsAnim3rb (3, Legs3RB, 1, _V(0.028,-0.646,0.176), _V(-0.1639,-0.9287,-0.3328),(float)(-PI/12.5));
	Parent5 = AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3rb);
	static UINT Legs3RG[1] = {20};
	static MGROUP_TRANSLATE LegsAnim3rg (3, Legs3RG, 1, _V(-0.12,0.018,0.00888));
	AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3rg, Parent5);


	//4.collapse
	static MGROUP_SCALE col1 (0, ArrayLFoldGrp, 1, _V(0,0,0), _V(1,1,0.05));
	static MGROUP_ROTATE col2 (0, ArrayLFoldGrp, 1, _V(0,0,-1.135), _V(-1,0,0), (float)(PI*0.47));
	anim_col = CreateAnimation (0);
	AddAnimationComponent (anim_col, 0, 0.75, &col2);
	AddAnimationComponent (anim_col, 0.75, 1, &col1);


	
}

void mpx::RevertAntenna (void)
{
	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void mpx::RevertArray (void)
{
	fold_status = ((fold_status == DOOR_CLOSED || fold_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}


void mpx::RevertLegs (void)
{
	legs_status = ((legs_status == DOOR_CLOSED || legs_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void mpx::RevertCol (void)
{
	col_status = ((col_status == DOOR_CLOSED || col_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

// ==============================================================
// mpx class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void mpx::clbkSetClassCaps (FILEHANDLE cfg)
{
	ant_proc = 0.0;
	ant_status = DOOR_CLOSED;
	fold_proc = 0.0;
	fold_status = DOOR_CLOSED;
	legs_proc = 0.0;
	legs_status = DOOR_CLOSED;
	col_proc = 0.0;
	col_status = DOOR_CLOSED;
	//rocksw = 0;
	texsw = 0;
	status = 0;
	is_attached = true;
	DefineAnimations ();
	// physical specs
	//SetCW (2.2, 2.2, 2, 2); old
	SetCW (2.0, 2.0, 2, 2);
	SetRotDrag (_V(2,2,2));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (1e-3);
		SetBankMomentScale (1e-3);
	}
	SetTrimScale (0.1);
	SetPMI (_V(0.4,0.4,0.3));
	SetCrossSections (_V(2,2,8));
	SetSize (2);

	hmpx_backshellmesh = oapiLoadMesh("Phoenix_Mars_Lander/mpx_backshell");
	SetMeshVisibilityMode ((AddMesh (hmpx_backshellmesh)), MESHVIS_ALWAYS);
	hmpx_heatshieldmesh = oapiLoadMesh("Phoenix_Mars_Lander/mpx_heatshield");
	SetMeshVisibilityMode ((AddMesh (hmpx_heatshieldmesh)), MESHVIS_ALWAYS);
	SetMeshVisibilityMode ((AddMesh ("Phoenix_Mars_Lander/mpx_solp_stowed")), MESHVIS_ALWAYS);
	SetMeshVisibilityMode ((AddMesh ("Phoenix_Mars_Lander/mpx_stowed")), MESHVIS_ALWAYS);
	bs2 = oapiLoadTexture ("Phoenix_Mars_Lander/backshell2.dds", true);
	bs4 = oapiLoadTexture ("Phoenix_Mars_Lander/heatshield4.dds", true);
	SetCameraOffset (_V(0,2,-3.5));
	SetLiftCoeffFunc (LiftCoeff);
	//SetTouchdownPoints (_V(0,1,-0.01), _V(-1,-1,-0.01), _V(1,-1,-0.01));
	SetTouchdownPoints(tdvtx_geardown,ntdvtx_geardown);
	SetDockParams (_V(0,0,0), _V(0,1,0), _V(0,0,-1));

	// reentry pstream
	PARTICLESTREAMSPEC rps = {
		0, 2.5, 20, 0, 0.001, 0.5, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 1e8, 5e15
	};
	AddReentryStream (&rps);



	// propellant resources
	
	rcspr = CreatePropellantResource (PA_FUELMASS, PA_FUELMASS);
	// ***************** thruster definitions *******************

	th_rcs[ 0] = CreateThruster (_V( -0.315,1.131, -0.1819), _V(-0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 1] = CreateThruster (_V( 0.315,1.131, -0.1819), _V(0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 2] = CreateThruster (_V( -0.315,-1.131, -0.1819), _V(-0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 3] = CreateThruster (_V( 0.315,-1.131, -0.1819), _V(0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);

	th_rcs[ 6] = CreateThruster (_V( -0.315,1.131, -0.1819), _V(-0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 7] = CreateThruster (_V( 0.315,1.131, -0.1819), _V(0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 8] = CreateThruster (_V( -0.315,-1.131, -0.1819), _V(-0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 9] = CreateThruster (_V( 0.315,-1.131, -0.1819), _V(0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);

	th_rcs[ 10] = CreateThruster (_V( -0.315,1.131, -0.1819), _V(-0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 11] = CreateThruster (_V( 0.315,1.131, -0.1819), _V(0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 12] = CreateThruster (_V( -0.315,-1.131, -0.1819), _V(-0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 13] = CreateThruster (_V( 0.315,-1.131, -0.1819), _V(0.5, 0,0.866), PB_MAXRCSTH, rcspr, PA_ISP);

	th_rcs[ 4] = CreateThruster (_V(0,0,-0.1819), _V(0, 0,1), PB_MAXRCSTH20, rcspr, PA_ISP);
	th_rcs[ 5] = CreateThruster (_V(0,0,0), _V(0, 0,1), 0, rcspr, PA_ISP);
	

	th_group[0] = th_rcs[2];
	th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[6];
	th_group[1] = th_rcs[9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[7];
	th_group[1] = th_rcs[8];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	//th_group[0] = th_rcs[5];
	//CreateThrusterGroup (th_group, 1, THGROUP_ATT_UP);

	//th_group[0] = th_rcs[5];
	//CreateThrusterGroup (th_group, 1, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[11];
	th_group[1] = th_rcs[13];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[10];
	th_group[1] = th_rcs[12];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	//th_group[0] = th_rcs[5];
	//CreateThrusterGroup (th_group, 1, THGROUP_ATT_LEFT);

	//th_group[0] = th_rcs[5];
	//CreateThrusterGroup (th_group, 1, THGROUP_ATT_RIGHT);

	//th_group[0] = th_rcs[5];
	//CreateThrusterGroup (th_group, 1, THGROUP_ATT_FORWARD);

	//th_group[0] = th_rcs[5];
	//CreateThrusterGroup (th_group, 1, THGROUP_ATT_BACK);

	CreateThrusterGroup (th_rcs+4, 1, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("exhaust_atrcs");
	AddExhaust (th_rcs[4], 1, 0.1, _V( -0.305,1.141, -0.2), _V(0,0,-1), (exhaust));
	AddExhaust (th_rcs[4], 1, 0.1, _V( 0.305,1.141, -0.2), _V(0,0,-1), (exhaust));
	AddExhaust (th_rcs[4], 1, 0.1, _V( -0.305,-1.141, -0.2), _V(0,0,-1), (exhaust));
	AddExhaust (th_rcs[4], 1, 0.1, _V( 0.305,-1.141, -0.2), _V(0,0,-1), (exhaust));

	AddExhaust (th_rcs[0], 0.5, 0.05, _V( -0.315,1.131, -0.21), _V(0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[1], 0.5, 0.05, _V( 0.315,1.131, -0.21), _V(-0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[2], 0.5, 0.05, _V( -0.315,-1.131, -0.21), _V(0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[3], 0.5, 0.05, _V( 0.315,-1.131, -0.21), _V(-0.5, 0,-0.866), (exhaust));

	AddExhaust (th_rcs[6], 0.5, 0.05, _V( -0.315,1.131, -0.21), _V(0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[7], 0.5, 0.05, _V( 0.315,1.131, -0.21), _V(-0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[8], 0.5, 0.05, _V( -0.315,-1.131, -0.21), _V(0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[9], 0.5, 0.05, _V( 0.315,-1.131, -0.21), _V(-0.5, 0,-0.866), (exhaust));

	AddExhaust (th_rcs[10], 0.5, 0.05, _V( -0.315,1.131, -0.21), _V(0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[11], 0.5, 0.05, _V( 0.315,1.131, -0.21), _V(-0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[12], 0.5, 0.05, _V( -0.315,-1.131, -0.21), _V(0.5, 0,-0.866), (exhaust));
	AddExhaust (th_rcs[13], 0.5, 0.05, _V( 0.315,-1.131, -0.21), _V(-0.5, 0,-0.866), (exhaust));

	//// attach point

	hAttach = CreateAttachment (false, _V( 0, 0, 0), _V( 0,0,1), _V(0,1,0), "LNDR");
	hDelta = CreateAttachment (true, _V( 0, 0, -1.3), _V( 0,0,-1), _V(0,1,0), "D2");

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new mpx (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (mpx*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void mpx::clbkPostCreation ()
{
	if(fold_status > 0)SetSize (10);
}

void mpx::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL2::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "CONFIGURATION", status);
	//oapiWriteScenario_int (scn, "ROCKS", rocksw);
	oapiWriteScenario_int (scn, "TEX", texsw);
	sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);
	sprintf (cbuf, "%d %0.4f", fold_status, fold_proc);
	oapiWriteScenario_string (scn, "CHUTES", cbuf);
	sprintf (cbuf, "%d %0.4f", legs_status, legs_proc);
	oapiWriteScenario_string (scn, "LEGS", cbuf);
	sprintf (cbuf, "%d %0.4f", col_status, col_proc);
	oapiWriteScenario_string (scn, "COL", cbuf);
}
void mpx::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "CONFIGURATION", 13)) {
            sscanf (line+13, "%d", &status);
		//} else if (!strnicmp (line, "ROCKS", 5)) {
        //    sscanf (line+5, "%d", &rocksw);
		} else if (!strnicmp (line, "TEX", 3)) {
            sscanf (line+3, "%d", &texsw);
		} else if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+4, "%d%lf", &ant_status, &ant_proc);
		} else if (!strnicmp (line, "CHUTES", 6)) {
			sscanf (line+6, "%d%lf", &fold_status, &fold_proc);
		} else if (!strnicmp (line, "LEGS", 4)) {
			sscanf (line+4, "%d%lf", &legs_status, &legs_proc);
		} else if (!strnicmp (line, "COL", 3)) {
			sscanf (line+3, "%d%lf", &col_status, &col_proc);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_ant, ant_proc);
	SetAnimation (anim_fold, fold_proc);
	SetAnimation (anim_legs, legs_proc);
	SetAnimation (anim_col, col_proc);
}
int mpx::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {
      case OAPI_KEY_J:  // jettison
		if (ant_status == 1)Jettison();
        return 1;
	case OAPI_KEY_K: // deploy solp
		if (ant_status == 0)
		{
			RevertAntenna();
			hmpx_landermesh = oapiLoadMeshGlobal ("Phoenix_Mars_Lander/mpx_lander");
		}
		return 1;
	case OAPI_KEY_G: // open/fold solar arrays
		RevertArray();
		return 1;
	case OAPI_KEY_DIVIDE: 
		RESETKEY (kstate, OAPI_KEY_LBRACKET);
		return 1;
	}
  }
  return 0;
}
void mpx::SpawnObject(char* classname, char* ext, VECTOR3 ofs, VECTOR3 vofs) {
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  oapiCreateVessel (ext, classname, vs);
}

void mpx::Jettison() 
{
  switch(status) {
    case 0:
      status=1;
      SpawnObject("Vessels/Phoenix_Mars_Lander/phoenix_solp","Phoenix_Cruise_Module",_V(0,0,-1.1),_V(0,0,-0.3));
      break;
  }
}

void mpx::Jettison2() {
  switch(status) {
    case 1:
      status=2;
      SpawnObject("Vessels/Phoenix_Mars_Lander/phoenix_heatshield","Phoenix_Heatshield",_V(0,0,0),_V(0,0,0.3));
      break;
  }
}

void mpx::Jettison3() {
  switch(status) {
    case 2:
      status=3;
      SpawnObject2("Vessels/Phoenix_Mars_Lander/phoenix_lander","Phoenix_Mars_Lander",_V(0,0,0),_V(0,0,0.0));
      break;
  }
}

void mpx::SpawnObject2(char* classname, char* ext, VECTOR3 ofs2, VECTOR3 vofs2)
{
	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(1,0,0), pos);
	GlobalRot (_V(0,1,0), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);


  VESSELSTATUS vs;
  //char name[256];
  GetStatus(vs);
  VECTOR3 rvel;
  VECTOR3 angvel = vs.vrot;
  Local2Rel (ofs2, vs.rpos);
  GlobalRot (vofs2, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  vs.vrot.x = vs.vrot.x;
  vs.vrot.y = -angvel.z;
  vs.vrot.z = angvel.y;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  double curr_fuel = GetPropellantMass(rcspr);
  OBJHANDLE hVes = oapiCreateVessel (ext, classname, vs);
  VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hVes);
  ATTACHMENTHANDLE hLndrAttach = v->GetAttachmentHandle(true, 0);
  AttachChild(hVes,hAttach,hLndrAttach);
  PROPELLANT_HANDLE hProp = v->GetPropellantHandleByIndex(0);
  v->SetPropellantMass(hProp, curr_fuel);
  //oapiSetFocusObject(hVes);
}

//void mpx::SpawnObject3(char* classname, char* ext) {
//
//	char name[256];
//  VESSELSTATUS2 vs;
//	memset (&vs, 0, sizeof(vs));
//	vs.version = 2;
//	vs.rbody = GetSurfaceRef();
//	vs.status = 1;
//	OBJHANDLE hVessel = GetHandle();
//	double longitude, latitude, radius;
//	double heading;
//	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
//	oapiGetHeading(hVessel,&heading);
//	vs.surf_lat = latitude;
//	vs.surf_lng = longitude;
//	vs.surf_hdg = heading;
//  oapiCreateVesselEx (ext, classname, &vs);
//}

double mpx::CalcEmptyMass() 
{
	double EmptyMass=0;
	if (status == 0) {
	EmptyMass+=MPX_SOLP_MASS + MPX_HTSHLD_MASS + MPX_LNDR_MASS;
	}

	if (status == 1) {
	EmptyMass+=MPX_HTSHLD_MASS + MPX_LNDR_MASS;
	}

	if (status == 2) {
	EmptyMass+=MPX_LNDR_MASS;
	}

	if (status == 3) {
	EmptyMass+=MPX_LNDR_MASS;
	}

	EmptyMass+=MPX_BKSHLL_MASS;
	return EmptyMass;
} 

void mpx::clbkPreStep(double SimT, double SimDT, double MJD) 
{
	if (is_attached == true)
	{
		OBJHANDLE hLauncher = GetAttachmentStatus(hDelta);
		if(hLauncher == NULL)is_attached = false;
	}


	SetEmptyMass(CalcEmptyMass());

if (is_attached == false)
	{
	double cwalt = GetAltitude(ALTMODE_MEANRAD);
	if (200000 > cwalt && 30000 < cwalt)
		{
		double hsa;
		hsa = ((cwalt-30000)*0.0001)+2.0;
		SetCW (hsa,hsa,2,2);
		}

	double alt_grnd = GetAltitude(ALTMODE_GROUND);

	if (13000 > alt_grnd && fold_status == 0 && status == 1) DoChutes();

	if (13000 > alt_grnd && fold_status > 0)
		{
		SetCrossSections (_V(8,8,(10+(fold_proc*40))));
	//SetCW (2.2, 2.2, 2, 2); old
		SetCW (2.0, 2.0, 1, 1);
	//SetRotDrag (_V(2,2,2));old
		SetRotDrag (_V(2.5,2.5,2.5));
		SetPitchMomentScale(0.0005);
		SetBankMomentScale(0.0005);
		SetClipRadius (10);
		SetPMI (_V(2,2,4));
		}

	if (12000 > alt_grnd && status == 1) Jettison2();

	if (11000 > alt_grnd && legs_status == 0 && status == 2) RevertLegs();

	if (10000 > alt_grnd && status == 2) Jettison3();

	if (800 > alt_grnd && status == 3)
		{
		DetachChild(hAttach,0.3);
		status = 4;
		}

	if (0.1 > alt_grnd) DoCol();
	

	if (35000 > alt_grnd && texsw == 0) 
		{
		ApplySkin();
		texsw = 1;
		}
	}

	////// animations
		// Animate hi-gain antenna
	if (ant_status >= DOOR_CLOSING) {
		double da = SimDT * ANTENNA_OPERATING_SPEED;
		if (ant_status == DOOR_CLOSING) {
			if (ant_proc > 0.0) ant_proc = max (0.0, ant_proc-da);
			else                ant_status = DOOR_CLOSED;
		} else {
			if (ant_proc < 1.0) ant_proc = min (1.0, ant_proc+da);
			else                ant_status = DOOR_OPEN;
		}
		SetAnimation (anim_ant, ant_proc);
	}

	// Animate solar arrays
	if (fold_status >= DOOR_CLOSING) {
		double da = SimDT * ARRAY_OPERATING_SPEED;
		if (fold_status == DOOR_CLOSING) {
			if (fold_proc > 0.0) fold_proc = max (0.0, fold_proc-da);
			else                 fold_status = DOOR_CLOSED;
		} else {
			if (fold_proc < 1.0) fold_proc = min (1.0, fold_proc+da);
			else                 fold_status = DOOR_OPEN;
		}
		SetAnimation (anim_fold, fold_proc);
	}

	// animate legs
	if (legs_status >= DOOR_CLOSING) {
		double da = SimDT * LEGS_OPERATING_SPEED;
		if (legs_status == DOOR_CLOSING) {
			if (legs_proc > 0.0) legs_proc = max (0.0, legs_proc-da);
			else                 legs_status = DOOR_CLOSED;
		} else {
			if (legs_proc < 1.0) legs_proc = min (1.0, legs_proc+da);
			else                 legs_status = DOOR_OPEN;
		}
		SetAnimation (anim_legs, legs_proc);
	}

	// animate collapse
	if (col_status >= DOOR_CLOSING) {
		double da = SimDT * COL_OPERATING_SPEED;
		if (col_status == DOOR_CLOSING) {
			if (col_proc > 0.0) col_proc = max (0.0, col_proc-da);
			else                 col_status = DOOR_CLOSED;
		} else {
			if (col_proc < 1.0) col_proc = min (1.0, col_proc+da);
			else                 col_status = DOOR_OPEN;
		}
		SetAnimation (anim_col, col_proc);
	}
}

void mpx::ApplySkin ()
{	
	 bs1 = oapiGetTextureHandle (hmpx_backshellmesh, 1);
	oapiBlt(bs1, bs2, 0, 0, 0, 0, 512, 512);

	bs3 = oapiGetTextureHandle (hmpx_heatshieldmesh, 2);
	oapiBlt(bs3, bs4, 0, 0, 0, 0, 512, 512);
}

void mpx::DoChutes() {

	RevertArray();
	SetSize (10);
}

void mpx::DoCol() {

	if (col_status == 0) RevertCol();
}

bool mpx::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);


	if (ant_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"K = Deploy Solar Panels");
	skp->Text ((10+roxl), (wc+royl), kbuf,strlen(kbuf));
	}

	
	else 
	{
	// get the moon
	char cbuf[256]="Mars";
	OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 marspos;
	GetRelativePos (hRef, marspos);
	if(length(marspos) < 4389000 && status == 0)
		{
		skp->SetTextColor(0x006666FF);
		char kbuf[256];
		sprintf(kbuf,"J = Jettison Cruise Module");
		skp->Text ((10+roxl), (wc+royl), kbuf,strlen(kbuf));
		}
	else if (status == 0)
		{
		skp->SetTextColor(0x0066FF66);
		char kbuf[256];
		sprintf(kbuf,"J = Jettison Cruise Module");
		skp->Text ((10+roxl), (wc+royl), kbuf,strlen(kbuf));
		}
	}
	
if (status < 3)
  {
	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(rcspr);
	double test = ((m1+m2-29.946)/m1);
	double dv = (2500*(log(test)));
	if (GetPropellantMass(rcspr) > 29.946)skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x006666FF);
	char kbuf[256];
	sprintf(kbuf,"Remaining (Cruise) dV %0.2f m/s", dv);
	skp->Text ((10+roxl), (w0+royl), kbuf,strlen(kbuf));
	}


	
	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(rcspr);
	double test = ((m1+m2)/m1);
	double dv = (2500*(log(test)));
	if (m2 < 29.946)skp->SetTextColor(0x006666FF);
	else skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"Remaining (Total) dV = %0.2f m/s", dv);
	skp->Text ((10+roxl), (w1+royl), kbuf,strlen(kbuf));
	}
	

	
	{
	if (GetPropellantMass(rcspr) < 29.946)skp->SetTextColor(0x006666FF);
	else skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"Remaining Fuel = %0.2f kg", GetPropellantMass(rcspr));
	skp->Text ((10+roxl), (w2+royl), kbuf,strlen(kbuf));
	}

	
	{
	if (GetPropellantMass(rcspr) < 30)skp->SetTextColor(0x006666FF);
	else skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"Lander Min.Fuel Req: 29.95kg");
	skp->Text ((10+roxl), (w3+royl), kbuf,strlen(kbuf));
	}
  }
	return true;
}


void mpx::clbkPostStep (double simt, double simdt, double mjd)
{
	if (status > 0) DelMesh(2);

	if (status > 1) DelMesh(1);

	if (status > 2) 
	{
		DelMesh(3);
		SetPropellantMass(rcspr, 0.0);
	}
	
}