// ==============================================================
//                 ORBITER MODULE: MARS PHOENIX LANDER
//
//
// mpxl.cpp
// Control module for MARS PHOENIX LANDER vessel class
//
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double PA_ISP = 2500;
const double PB_MAXRCSTH = 4.5;
const double PB_MAXRCSTH20 = 1211;
const double PB_LID = 0;
const double MPX_LNDR_MASS = 424;
const double ANTENNA_OPERATING_SPEED = 0.1;
const double ARRAY_OPERATING_SPEED = 0.025;
const double LEGS_OPERATING_SPEED = 1;
const double FEET_OPERATING_SPEED = 10;
const double COV_OPERATING_SPEED = 0.25;
const double ARMROT_SPEED = 0.1;
const double ARMSLD_SPEED = 0.1;
const double PA_FUELMASS = 62;

static const DWORD ntdvtx_geardown = 4;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(   0,-0.60, 2), 1e5, 1e4, 0.5, 0.5},
	{_V(-1.732,-0.60,-1), 1e5, 1e4, 0.5, 0.5},
	{_V( 1.732,-0.60,-1), 1e5, 1e4, 0.5, 0.5},
	{_V(   0,  1,   0), 1e5, 1e4, 0.5}
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}


class mpxl: public VESSEL3 {
public:
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } ant_status, legs_status, feet_status, cov_status;
	mpxl (OBJHANDLE hVessel, int flightmodel);
	~mpxl ();
	
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);

	private:

	void DefineAnimations (void);
	void RevertAntenna (void);
	void RevertLegs (void);
	void RevertFeet (void);
	void RevertCov (void);
	void DoDescent (void);
	void DoOrient (void);
	void DoAttitude (void);
	void DoCrash (void);
	void AddSmoke(void);
	void SpawnRocks(void);
	MATRIX3 RotationMatrix(VECTOR3 angles, bool xyz = FALSE);

	THRUSTER_HANDLE th_rcs[7], th_group[4], th_lid;
	PROPELLANT_HANDLE rcspr, ph_lid;
	NOTEHANDLE hNote;
	ANIMATIONCOMPONENT_HANDLE Parent, Parent1, Parent2, Parent3, Parent4, Parent5, Parent6, Parent7, Parent8;
	ANIMATIONCOMPONENT_HANDLE Parent9, Parent10, Parent11;
	ANIMATIONCOMPONENT_HANDLE p1, p2, p3, p4;
	ANIMATIONCOMPONENT_HANDLE a1, a2, a3, a4;

	int crashsw;
	int legsw;
	int feetsw;
	int thsw;
	int thsw2;
	int thsw3;
	int armrotsw;
	int armsldsw;
	int armelbsw;
	int hdsw;
	int lidsw;
	int strb;
	int manov;
	int camrot;
	int smoke_sw;
	double half;
	double crtime;
	double hdg;
	int rock_sw;

	UINT anim_ant, anim_fold, anim_legs, anim_feet, anim_cov, anim_arp, anim_sld, anim_elb;
	double ant_proc, fold_proc, legs_proc, feet_proc, cov_proc, arp_proc, sld_proc, elb_proc;
};

mpxl::mpxl (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

mpxl::~mpxl ()
{

}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void mpxl::DefineAnimations (void)
{
	// 1. solp
	anim_ant = CreateAnimation (0);
	static UINT solpl1[2] = {14,51};
	static MGROUP_ROTATE HiGainAntl1 (0, solpl1, 2, _V(-0.789,0,0.356), _V(0,1,0), (float)(PI/2));
	p1 = AddAnimationComponent (anim_ant, 0, 0.05, &HiGainAntl1);

	static UINT solpl2[1] = {52};
	static MGROUP_ROTATE HiGainAntl2 (0, solpl2, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI/2));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl2, p1);

	static UINT solpl3[1] = {76};
	static MGROUP_ROTATE HiGainAntl3 (0, solpl3, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl3, p2);

	static UINT solpl4[1] = {58};
	static MGROUP_ROTATE HiGainAntl4 (0, solpl4, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl4, p1);

	static UINT solpl5[1] = {75};
	static MGROUP_ROTATE HiGainAntl5 (0, solpl5, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl5, p2);

	static UINT solpl6[1] = {59};
	static MGROUP_ROTATE HiGainAntl6 (0, solpl6, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl6, p1);

	static UINT solpl7[1] = {74};
	static MGROUP_ROTATE HiGainAntl7 (0, solpl7, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl7, p2);

	static UINT solpl8[1] = {60};
	static MGROUP_ROTATE HiGainAntl8 (0, solpl8, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl8, p1);

	static UINT solpl9[1] = {73};
	static MGROUP_ROTATE HiGainAntl9 (0, solpl9, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl9, p2);

	static UINT solpl10[1] = {61};
	static MGROUP_ROTATE HiGainAntl10 (0, solpl10, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl10, p1);

	static UINT solpl11[1] = {72};
	static MGROUP_ROTATE HiGainAntl11 (0, solpl11, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl11, p2);

	static UINT solpl12[1] = {62};
	static MGROUP_ROTATE HiGainAntl12 (0, solpl12, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl12, p1);

	static UINT solpl13[1] = {71};
	static MGROUP_ROTATE HiGainAntl13 (0, solpl13, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl13, p2);

	static UINT solpl14[1] = {63};
	static MGROUP_ROTATE HiGainAntl14 (0, solpl14, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl14, p1);

	static UINT solpl15[1] = {70};
	static MGROUP_ROTATE HiGainAntl15 (0, solpl15, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl15, p2);

	static UINT solpl16[1] = {64};
	static MGROUP_ROTATE HiGainAntl16 (0, solpl16, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl16, p1);

	static UINT solpl17[1] = {69};
	static MGROUP_ROTATE HiGainAntl17 (0, solpl17, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl17, p2);

	static UINT solpl18[1] = {65};
	static MGROUP_ROTATE HiGainAntl18 (0, solpl18, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl18, p1);

	static UINT solpl19[1] = {68};
	static MGROUP_ROTATE HiGainAntl19 (0, solpl19, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl19, p2);

	static UINT solpl20[1] = {66};
	static MGROUP_ROTATE HiGainAntl20 (0, solpl20, 1, _V(-0.7755,0.275,-0.556), _V(0,0,1), (float)(PI));
	p2 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl20, p1);

	static UINT solpl21[1] = {67};
	static MGROUP_ROTATE HiGainAntl21 (0, solpl21, 1, _V(-0.7755,0.275,-0.556), _V(0,0.309,-0.951), (float)(PI));
	p1 = AddAnimationComponent (anim_ant, 0.1, 0.75, &HiGainAntl21, p2);




	//solp right
	static UINT solpr1[2] = {13,50};
	static MGROUP_ROTATE HiGainAntr1 (0, solpr1, 2, _V(0.789,0,0.356), _V(0,-1,0), (float)(PI/2));
	p3 = AddAnimationComponent (anim_ant, 0.05, 0.1,   &HiGainAntr1);

	static UINT solpr2[1] = {32};
	static MGROUP_ROTATE HiGainAntr2 (0, solpr2, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI/2));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr2, p3);

	static UINT solpr3[1] = {95};
	static MGROUP_ROTATE HiGainAntr3 (0, solpr3, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr3, p4);

	static UINT solpr4[1] = {77};
	static MGROUP_ROTATE HiGainAntr4 (0, solpr4, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr4, p3);

	static UINT solpr5[1] = {94};
	static MGROUP_ROTATE HiGainAntr5 (0, solpr5, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr5, p4);

	static UINT solpr6[1] = {78};
	static MGROUP_ROTATE HiGainAntr6 (0, solpr6, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr6, p3);

	static UINT solpr7[1] = {93};
	static MGROUP_ROTATE HiGainAntr7 (0, solpr7, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr7, p4);

	static UINT solpr8[1] = {79};
	static MGROUP_ROTATE HiGainAntr8 (0, solpr8, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr8, p3);

	static UINT solpr9[1] = {92};
	static MGROUP_ROTATE HiGainAntr9 (0, solpr9, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr9, p4);

	static UINT solpr10[1] = {80};
	static MGROUP_ROTATE HiGainAntr10 (0, solpr10, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr10, p3);

	static UINT solpr11[1] = {91};
	static MGROUP_ROTATE HiGainAntr11 (0, solpr11, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr11, p4);

	static UINT solpr12[1] = {81};
	static MGROUP_ROTATE HiGainAntr12 (0, solpr12, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr12, p3);

	static UINT solpr13[1] = {90};
	static MGROUP_ROTATE HiGainAntr13 (0, solpr13, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr13, p4);

	static UINT solpr14[1] = {82};
	static MGROUP_ROTATE HiGainAntr14 (0, solpr14, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr14, p3);

	static UINT solpr15[1] = {89};
	static MGROUP_ROTATE HiGainAntr15 (0, solpr15, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr15, p4);

	static UINT solpr16[1] = {83};
	static MGROUP_ROTATE HiGainAntr16 (0, solpr16, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr16, p3);

	static UINT solpr17[1] = {88};
	static MGROUP_ROTATE HiGainAntr17 (0, solpr17, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr17, p4);

	static UINT solpr18[1] = {84};
	static MGROUP_ROTATE HiGainAntr18 (0, solpr18, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr18, p3);

	static UINT solpr19[1] = {87};
	static MGROUP_ROTATE HiGainAntr19 (0, solpr19, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr19, p4);

	static UINT solpr20[1] = {85};
	static MGROUP_ROTATE HiGainAntr20 (0, solpr20, 1, _V(0.7755,0.275,-0.556), _V(0,0,-1), (float)(PI));
	p4 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr20, p3);

	static UINT solpr21[1] = {86};
	static MGROUP_ROTATE HiGainAntr21 (0, solpr21, 1, _V(0.7755,0.275,-0.556), _V(0,-0.309,0.951), (float)(PI));
	p3 = AddAnimationComponent (anim_ant, 0.35, 1, &HiGainAntr21, p4);







	// 2. chutes
	anim_fold = CreateAnimation (0);
	static UINT ArrayLFoldGrp[1] = {0};
	static UINT ArrayRFoldGrp[1] = {5};

	static MGROUP_SCALE ArrayLFold1 (0, ArrayLFoldGrp, 1, _V(0,0,-0.8), _V(12.5,12.5,150));
	AddAnimationComponent (anim_fold, 0,   0.5, &ArrayLFold1);
	static MGROUP_SCALE ArrayLFold2 (0, ArrayLFoldGrp, 1, _V(0,0,-0.8), _V(12,12,1));
	AddAnimationComponent (anim_fold, 0.5,   1, &ArrayLFold2);
	static MGROUP_SCALE ArrayLFold3 (0, ArrayRFoldGrp, 1, _V(0,0,-1), _V(0.01,0.01,0.01));
	AddAnimationComponent (anim_fold, 0,   0.2, &ArrayLFold3);


	// 3. legs
	anim_legs = CreateAnimation (0);
	static UINT Legs1[1] = {99};
	static MGROUP_TRANSLATE LegsAnim1 (0, Legs1, 1, _V(0,-0.24941376,0.144));
	Parent6 = AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1);

	static UINT Legs1LB[1] = {48};
	static MGROUP_ROTATE LegsAnim1lb (0, Legs1LB, 1, _V(-0.574,-0.176,0.299), _V(0.7165,-0.3488,-0.6041),(float)(PI/12.5));
	Parent = AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1lb);
	static UINT Legs1LG[1] = {47};
	static MGROUP_TRANSLATE LegsAnim1lg (0, Legs1LG, 1, _V(0.0756,-0.00888,0.0948));
	AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1lg, Parent);

	static UINT Legs1RB[1] = {37};
	static MGROUP_ROTATE LegsAnim1rb (0, Legs1RB, 1, _V(0.574,-0.176,0.299), _V(-0.7165,-0.3488,-0.6041),(float)(-PI/12.5));
	Parent1 = AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1rb);
	static UINT Legs1RG[1] = {46};
	static MGROUP_TRANSLATE LegsAnim1rg (0, Legs1RG, 1, _V(-0.0756,-0.00888,0.0948));
	AddAnimationComponent (anim_legs, 0, 0.333, &LegsAnim1rg, Parent1);

	//feet
	anim_feet = CreateAnimation (0);
	static UINT Legs1ft[2] = {10,55};
	static MGROUP_ROTATE LegsAnim1ft (0, Legs1ft, 2, _V(0,-0.25,0.995), _V(1,0,0),(float)(PI*0.1055));
	AddAnimationComponent (anim_feet, 0, 1, &LegsAnim1ft, Parent6);





	static UINT Legs2[1] = {100};
	static MGROUP_TRANSLATE LegsAnim2 (0, Legs2, 1, _V(0.12470688,-0.24941376,-0.072));
	Parent7 = AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2);

	static UINT Legs2LB[1] = {39};
	static MGROUP_ROTATE LegsAnim2lb (0, Legs2LB, 1, _V(0.545,-0.176,0.348), _V(-0.8863,-0.3329,-0.322),(float)(PI/12.5));
	Parent2 = AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2lb);
	static UINT Legs2LG[1] = {43};
	static MGROUP_TRANSLATE LegsAnim2lg (0, Legs2LG, 1, _V(0.0444,-0.00888,-0.11304));
	AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2lg, Parent2);

	static UINT Legs2RB[1] = {38};
	static MGROUP_ROTATE LegsAnim2rb (0, Legs2RB, 1, _V(-0.028,-0.176,-0.646), _V(0.1639,0.3328,-0.9287),(float)(PI/12.5));
	Parent3 = AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2rb);
	static UINT Legs2RG[1] = {42};
	static MGROUP_TRANSLATE LegsAnim2rg (0, Legs2RG, 1, _V(0.12,-0.00888,0.018));
	AddAnimationComponent (anim_legs, 0.333, 0.666, &LegsAnim2rg, Parent3);

	//feet
	static UINT Legs2ft[2] = {11,56};
	static MGROUP_ROTATE LegsAnim2ft (0, Legs2ft, 2, _V(0.862,-0.25,-0.4975), _V(0.5,0,0.866),(float)(-PI*0.1055));
	AddAnimationComponent (anim_feet, 0, 1, &LegsAnim2ft, Parent7);




	static UINT Legs3[1] = {98};
	static MGROUP_TRANSLATE LegsAnim3 (0, Legs3, 1, _V(-0.12470688,-0.24941376,-0.072));
	Parent8 = AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3);


	static UINT Legs3LB[1] = {40};
	static MGROUP_ROTATE LegsAnim3lb (0, Legs3LB, 1, _V(-0.545,-0.176,0.348), _V(0.8863,-0.3329,-0.322),(float)(-PI/12.5));
	Parent4 = AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3lb);
	static UINT Legs3LG[1] = {44};
	static MGROUP_TRANSLATE LegsAnim3lg (0, Legs3LG, 1, _V(-0.0444,-0.00888,-0.11304));
	AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3lg, Parent4);

	static UINT Legs3RB[1] = {41};
	static MGROUP_ROTATE LegsAnim3rb (0, Legs3RB, 1, _V(0.028,-0.176,-0.646), _V(-0.1639,0.3328,-0.9287),(float)(-PI/12.5));
	Parent5 = AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3rb);
	static UINT Legs3RG[1] = {45};
	static MGROUP_TRANSLATE LegsAnim3rg (0, Legs3RG, 1, _V(-0.12,-0.00888,0.018));
	AddAnimationComponent (anim_legs, 0.666, 1, &LegsAnim3rg, Parent5);

	//feet
	static UINT Legs3ft[2] = {12,57};
	static MGROUP_ROTATE LegsAnim3ft (0, Legs3ft, 2, _V(-0.862,-0.25,-0.4975), _V(0.5,0,-0.866),(float)(-PI*0.1055));
	AddAnimationComponent (anim_feet, 0, 1, &LegsAnim3ft, Parent8);

	// 5. cover
	anim_cov = CreateAnimation (0);
	static UINT covgrp1[1] = {36};
	static MGROUP_ROTATE cover1 (0, covgrp1, 1, _V(0.49,0.277,0.289), _V(0.7071,0,0.7071), (float)(PI/8));
	Parent9 = AddAnimationComponent (anim_cov, 0, 0.25, &cover1);

	static UINT covgrp2[1] = {97};
	static MGROUP_ROTATE cover2 (0, covgrp2, 1, _V(0.49,0.277,0.289), _V(0.7071,0,0.7071), (float)(PI/4));
	Parent10 = AddAnimationComponent (anim_cov, 0, 0.25, &cover2, Parent9);

	static UINT covgrp3[1] = {96};
	static MGROUP_ROTATE cover3 (0, covgrp3, 1, _V(0.49,0.277,0.289), _V(0.7071,0,0.7071), (float)(PI/4));
	Parent11 = AddAnimationComponent (anim_cov, 0, 0.25, &cover3, Parent10);

	static UINT covgrp4[1] = {35};
	static MGROUP_ROTATE cover4 (0, covgrp4, 1, _V(0.49,0.277,0.289), _V(0.7071,0,0.7071), (float)(PI/4));
	AddAnimationComponent (anim_cov, 0, 0.25, &cover4, Parent11);

	static UINT covgrp5[2] = {2,3};
	static MGROUP_ROTATE cover5 (0, covgrp5, 2, _V(0.47, 0.309, 0.603), _V(-0.7071,0,0.7071), (float)(-PI/2));
	AddAnimationComponent (anim_cov, 0.1, 0.35, &cover5);

	static UINT covgrp6[1] = {19};
	static MGROUP_ROTATE cover6 (0, covgrp6, 1, _V(-0.163,0.609,0), _V(0,0,1), (float)(-5*PI/4));
	AddAnimationComponent (anim_cov, 0.35, 0.4, &cover6);

	static UINT camgrp1[4] = {6,7,8,9};
	static MGROUP_TRANSLATE cam1 (0, camgrp1, 4, _V(0,0.6125,0));
	AddAnimationComponent (anim_cov, 0.5, 1, &cam1);

	static UINT camgrp2[1] = {20};
	static MGROUP_SCALE cam2 (0, camgrp2, 1, _V(-0.505,0.293,-0.087), _V(1,17.85625,1));
	AddAnimationComponent (anim_cov, 0.5, 1, &cam2);

	// 2. camera rot
	anim_fold = CreateAnimation (0);
	static UINT camgrp3[4] = {6,7,8,9};
	static MGROUP_ROTATE cam3 (0, camgrp3, 4, _V(-0.505,0.293,-0.087), _V(0,1,0),(float)(2*PI));
	AddAnimationComponent (anim_fold, 0, 1, &cam3);

	// arm rot pos
	anim_arp = CreateAnimation (0.7);
	static UINT armgrp1[1] = {27};
	static MGROUP_ROTATE arp1 (0, armgrp1, 1, _V(-0.502,0,-0.641), _V(0,1,0),(float)(0.83333*PI));
	a1 = AddAnimationComponent (anim_arp, 0, 1, &arp1);

	// arm sld pos
	anim_sld = CreateAnimation (0);
	static UINT sldgrp1[3] = {21,22,30};
	static MGROUP_ROTATE sld1 (0, sldgrp1, 3, _V(-0.42,0.327,-0.723), _V(0.7071,0,-0.7071),(float)(-PI));
	a2 = AddAnimationComponent (anim_sld, 0, 1, &sld1, a1);

	static UINT sldgrp2[4] = {23,25,26,29};
	static MGROUP_ROTATE sld2 (0, sldgrp2, 0, _V(0.367,0.327,0.274), _V(0.7071,0,-0.7071),(float)(0.6*PI));
	a3 = AddAnimationComponent (anim_sld, 0, 1, &sld2, a2);

	anim_elb = CreateAnimation (0);
	static UINT elbgrp1[4] = {23,25,26,29};
	static MGROUP_ROTATE elb1 (0, elbgrp1, 4, _V(0.367,0.327,0.274), _V(0.7071,0,-0.7071),(float)(-0.5*PI));
	AddAnimationComponent (anim_elb, 0, 1, &elb1, a3);







}

void mpxl::RevertAntenna (void)
{
	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void mpxl::RevertLegs (void)
{
	legs_status = ((legs_status == DOOR_CLOSED || legs_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void mpxl::RevertFeet (void)
{
	feet_status = ((feet_status == DOOR_CLOSED || feet_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void mpxl::RevertCov (void)
{
	cov_status = ((cov_status == DOOR_CLOSED || cov_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

// ==============================================================
// mpxl class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void mpxl::clbkSetClassCaps (FILEHANDLE cfg)
{
	ant_proc = 0.0;
	ant_status = DOOR_CLOSED;
	arp_proc = 0.7;
	sld_proc = 0.0;
	elb_proc = 0.0;
	fold_proc = 0.0;
	legs_proc = 0.0;
	legs_status = DOOR_CLOSED;
	feet_proc = 0.0;
	feet_status = DOOR_CLOSED;
	cov_proc = 0.0;
	cov_status = DOOR_CLOSED;

	crashsw = 0;
	legsw = 0;
	feetsw = 0;
	thsw = 0;
	thsw2 = 0;
	thsw3 = 0;
	armrotsw = 0;
	armsldsw = 0;
	armelbsw = 0;
	hdsw = 0;
	lidsw = 0;
	strb = 0;
	manov = 0;
	camrot = 0;
	smoke_sw = 0;
	half = 0;
	crtime = 0;
	hdg = -1;
	rock_sw = 0;

	DefineAnimations ();
	// physical specs
	SetCW (2, 2, 0.9, 0.9);
	SetRotDrag (_V(1,1,1));
	SetSurfaceFrictionCoeff(1,1);
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (1e-3);
		SetBankMomentScale (1e-3);
	}
	SetTrimScale (0.1);
	SetPMI (_V(0.3,0.3,0.2));
	SetCrossSections (_V(2,2,6));
	SetSize (2);
	SetEmptyMass (MPX_LNDR_MASS);
	SetMeshVisibilityMode ((AddMesh ("Phoenix_Mars_Lander/mpx_lander")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(0,1,-2));
	SetLiftCoeffFunc (LiftCoeff);
	//SetTouchdownPoints (_V(0,-0.58, 1), _V(-1, -0.58, -1), _V(1, -0.58, -1));
	SetTouchdownPoints (tdvtx_geardown, ntdvtx_geardown);


	// reentry pstream
	/*PARTICLESTREAMSPEC rps = {
		0, 2.5, 40, 0, 0.001, 0.5, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 2e6, 5e15
	};*/
	//AddReentryStream (&rps);



	// propellant resources
	
	
	rcspr = CreatePropellantResource (PA_FUELMASS, PA_FUELMASS);
	ph_lid = CreatePropellantResource (1, 1);
	// ***************** thruster definitions *******************
	
	

	SURFHANDLE lidar = oapiRegisterExhaustTexture ("Phoenix_Mars_Lander/lidar");
	th_lid	= CreateThruster (_V(-0.213,0.619,0.24), _V(0, -1, 0), PB_LID, ph_lid, PA_ISP);
	AddExhaust (th_lid, 10, 0.01, lidar);

	th_rcs[ 0] = CreateThruster (_V( -0.53,0.262, 0.886), _V(-0.5, -0.866, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 1] = CreateThruster (_V( -0.282,0.262, 1.012), _V(0.5, -0.866, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 2] = CreateThruster (_V( 0.53,0.262, -0.886), _V(0.5, -0.866, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 3] = CreateThruster (_V( 0.282,0.262, -1.012), _V(-0.5, -0.866, 0), PB_MAXRCSTH, rcspr, PA_ISP);

	th_rcs[ 4] = CreateThruster (_V(0,-0.33,0.645), _V(0, 1,0), PB_MAXRCSTH20, rcspr, PA_ISP);
	th_rcs[ 5] = CreateThruster (_V(-0.5586,-0.33,-0.3225), _V(0, 1,0), PB_MAXRCSTH20, rcspr, PA_ISP);
	th_rcs[ 6] = CreateThruster (_V(0.5586,-0.33,-0.3225), _V(0, 1,0), PB_MAXRCSTH20, rcspr, PA_ISP);


	th_group[0] = th_rcs[4];
	//th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[5];
	th_group[1] = th_rcs[6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[6];
	//th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[5];
	//th_group[1] = th_rcs[2];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_BANKRIGHT);


	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);


	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[6];
	CreateThrusterGroup (th_group, 3, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("exhaust_atrcs");
	SURFHANDLE exhaust2 = oapiRegisterExhaustTexture ("Exhaust2");
	AddExhaust (th_rcs[4], 0.75, 0.075, _V(0.205,-0.34,0.558), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[4], 0.75, 0.075, _V(0.11,-0.34,0.613), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[4], 0.75, 0.075, _V(-0.205,-0.34,0.558), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[4], 0.75, 0.075, _V(-0.11,-0.34,0.613), _V(0,-1,0), (exhaust2));

	AddExhaust (th_rcs[5], 0.75, 0.075, _V(-0.38,-0.34,-0.454), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[5], 0.75, 0.075, _V(-0.476,-0.34,-0.399), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[5], 0.75, 0.075, _V(-0.585,-0.34,-0.21), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[5], 0.75, 0.075, _V(-0.585,-0.34,-0.099), _V(0,-1,0), (exhaust2));

	AddExhaust (th_rcs[6], 0.75, 0.075, _V(0.38,-0.34,-0.454), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[6], 0.75, 0.075, _V(0.476,-0.34,-0.399), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[6], 0.75, 0.075, _V(0.585,-0.34,-0.21), _V(0,-1,0), (exhaust2));
	AddExhaust (th_rcs[6], 0.75, 0.075, _V(0.585,-0.34,-0.099), _V(0,-1,0), (exhaust2));

	AddExhaust (th_rcs[0], 0.5, 0.05,  (exhaust));
	AddExhaust (th_rcs[1], 0.5, 0.05,  (exhaust));
	AddExhaust (th_rcs[2], 0.5, 0.05,  (exhaust));
	AddExhaust (th_rcs[3], 0.5, 0.05,  (exhaust));

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new mpxl (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (mpxl*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void mpxl::clbkPostCreation ()
{
	if(oapiSurfaceElevation (GetSurfaceRef(),0,0) == 0) rock_sw = 0;
	else rock_sw = 1;


		//VECTOR3 gpos, grms;
		//Local2Global (_V(0,0,0), grms);  // global position of lander

		//// Search the complete vessel list for a grappling candidate.
		//// Not very scalable ...
		//for (DWORD i = 0; i < oapiGetVesselCount(); i++)
		//	{
		//	OBJHANDLE hV = oapiGetVesselByIndex (i);
		//	if (hV == GetHandle()) continue; // we don't want to grapple ourselves ...
		//	oapiGetGlobalPos (hV, &gpos);
		//		if (dist (gpos, grms) < (oapiGetSize (hV)))
		//		{ // in range
		//		double pmass = oapiGetFuelMass (hV);
		//		SetPropellantMass (rcspr, pmass);

		//		return;
		//		}
		//	}
}


void mpxl::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL2::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "MANUAL", manov);

	oapiWriteScenario_int (scn, "ROCKS", rock_sw);

	sprintf (cbuf, "%0.4f", hdg);
	oapiWriteScenario_string (scn, "HEDG", cbuf);

	sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);

	sprintf (cbuf, "%d %0.4f", camrot, fold_proc);
	oapiWriteScenario_string (scn, "CAMERA", cbuf);

	sprintf (cbuf, "%d %0.4f", legs_status, legs_proc);
	oapiWriteScenario_string (scn, "LEGS", cbuf);

	sprintf (cbuf, "%d %0.4f", cov_status, cov_proc);
	oapiWriteScenario_string (scn, "COV", cbuf);

	sprintf (cbuf, "%0.4f", arp_proc);
	oapiWriteScenario_string (scn, "ARMROT", cbuf);

	sprintf (cbuf, "%0.4f", sld_proc);
	oapiWriteScenario_string (scn, "ARMSLD", cbuf);

	sprintf (cbuf, "%0.4f", elb_proc);
	oapiWriteScenario_string (scn, "ARMELB", cbuf);

	sprintf (cbuf, "%d %0.4f", feet_status, feet_proc);
	oapiWriteScenario_string (scn, "FEET", cbuf);
}
void mpxl::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "MANUAL", 6)) {
            sscanf (line+6, "%d", &manov);
		} else if (!strnicmp (line, "ROCKS", 5)) {
			sscanf (line+5, "%d", &rock_sw);
		} else if (!strnicmp (line, "HEDG", 4)) {
			sscanf (line+4, "%lf", &hdg);
		} else if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+4, "%d%lf", &ant_status, &ant_proc);
		} else if (!strnicmp (line, "CAMERA", 6)) {
			sscanf (line+6, "%d%lf", &camrot, &fold_proc);
		} else if (!strnicmp (line, "LEGS", 4)) {
			sscanf (line+4, "%d%lf", &legs_status, &legs_proc);
		} else if (!strnicmp (line, "COV", 3)) {
			sscanf (line+3, "%d%lf", &cov_status, &cov_proc);
		} else if (!strnicmp (line, "ARMROT", 6)) {
			sscanf (line+6, "%lf", &arp_proc);
		} else if (!strnicmp (line, "ARMSLD", 6)) {
			sscanf (line+6, "%lf", &sld_proc);
		} else if (!strnicmp (line, "ARMELB", 6)) {
			sscanf (line+6, "%lf", &elb_proc);
		} else if (!strnicmp (line, "FEET", 4)) {
			sscanf (line+4, "%d%lf", &feet_status, &feet_proc);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_ant, ant_proc);
	SetAnimation (anim_fold, fold_proc);
	SetAnimation (anim_legs, legs_proc);
	SetAnimation (anim_cov, cov_proc);
	SetAnimation (anim_feet, feet_proc);
	SetAnimation (anim_arp, arp_proc);
	SetAnimation (anim_sld, sld_proc);
	SetAnimation (anim_elb, elb_proc);
}
int mpxl::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {
	case OAPI_KEY_M: //manual override
		if (manov == 0) manov = 1;
		else if (manov == 1) manov = 0;
		return 1;
	case OAPI_KEY_D: //lidar
		if (cov_status == 1){
		if (lidsw == 0) lidsw = 1;
		else if (lidsw == 1) lidsw = 0;}
		return 1;
	case OAPI_KEY_G: // deploy ant, covers
		if (cov_status == 0 && GroundContact()) RevertCov();
		return 1;
	case OAPI_KEY_K: // deploy solp
		if (ant_status == 0 && GroundContact()) RevertAntenna();
		return 1;
	case OAPI_KEY_C: // camera rot
		if (cov_status == 1){
		if (camrot == 0) camrot = 1;
		else if (camrot == 1) camrot = 0;}
		return 1;
	case OAPI_KEY_NUMPAD5:
		if (manov == 0) RESETKEY (kstate, OAPI_KEY_NUMPAD5);
		else ToggleNavmode(1);
		return 1;
	case OAPI_KEY_L:
		if (manov == 0) RESETKEY (kstate, OAPI_KEY_L);
		else ToggleNavmode(2);
		return 1;
	case OAPI_KEY_LBRACKET:
		if (manov == 0) RESETKEY (kstate, OAPI_KEY_LBRACKET);
		else ToggleNavmode(3);
		return 1;
	case OAPI_KEY_RBRACKET:
		if (manov == 0) RESETKEY (kstate, OAPI_KEY_RBRACKET);
		else ToggleNavmode(4);
		return 1;
	case OAPI_KEY_SEMICOLON:
		if (manov == 0) RESETKEY (kstate, OAPI_KEY_SEMICOLON);
		else ToggleNavmode(5);
		return 1;
	case OAPI_KEY_APOSTROPHE:
		if (manov == 0) RESETKEY (kstate, OAPI_KEY_APOSTROPHE);
		else ToggleNavmode(6);
		return 1;
	}
  }
  return 0;
}

int mpxl::clbkConsumeDirectKey(char *kstate)
{
 if (KEYMOD_SHIFT (kstate))
 {
		if (KEYDOWN (kstate, OAPI_KEY_V))
		{
		armrotsw = 2;
		}

		if (KEYDOWN (kstate, OAPI_KEY_B))
		{
		armsldsw = 2;
		}

		if (KEYDOWN (kstate, OAPI_KEY_N))
		{
		armelbsw = 2;
		}
 }

 else{
	if (KEYDOWN (kstate, OAPI_KEY_V))
		{
		armrotsw = 1;
		}

	if (KEYDOWN (kstate, OAPI_KEY_B))
		{
		armsldsw = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_N))
		{
		armelbsw = 1;
		}

if (manov == 0)
	{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD4);
		}
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD6);
		}
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD2);
		}
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD8);
		}
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD1);
		}
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD3);
		}
if (KEYDOWN (kstate, OAPI_KEY_ADD))
		{
		RESETKEY (kstate, OAPI_KEY_ADD);
		}
if (KEYDOWN (kstate, OAPI_KEY_SUBTRACT))
		{
		RESETKEY (kstate, OAPI_KEY_SUBTRACT);
		}
if (KEYDOWN (kstate, OAPI_KEY_MULTIPLY))
		{
		RESETKEY (kstate, OAPI_KEY_MULTIPLY);
		}
	}
 }
return 0;
}


void mpxl::DoCrash (void)
{

if (crashsw == 0)
	{
crtime = oapiGetSimTime();
crashsw = 1;

VECTOR3 col = _V(1,0,0);
hNote = oapiCreateAnnotation(true, 2, col);
oapiAnnotationSetPos(hNote,0.3,0.4,0.8,0.8);
oapiAnnotationSetText(hNote, "Crash! Impact > 10m/s");
	}

}


void mpxl::DoAttitude (void)
{
	double mass = GetMass();
	double alt = GetAltitude(ALTMODE_GROUND);
	VECTOR3 aspdv;
	GetGroundspeedVector(FRAME_LOCAL, aspdv);
	VECTOR3 avel;
	GetAngularVel(avel);
	VECTOR3 weightv;
	double weight;
	GetWeightVector(weightv);
	weight = length(weightv);
	VECTOR3 haspdv;
	GetGroundspeedVector(FRAME_HORIZON, haspdv);

	double torquez = 1211*0.5586;
	double torquex = 1211*0.645;

	double pitch = GetPitch();
	double pf = (1/(cos(pitch)));

	double bank = GetBank();
	double bf = (1/(cos(bank)));

	double inrtx = (0.4*mass);
	double inrtz = (0.3*mass);

	if(hdg == -1)
	{
		hdg = atan2(haspdv.x, haspdv.z);
	}

	double reqpx = -atan(aspdv.z/aspdv.y);
	double reqpz = atan(aspdv.x/aspdv.y);

	if (alt > 550)
	{
	VECTOR3 bam_vector = _V(sin(hdg), -1.0, cos(hdg));
	VECTOR3 bam_vct_loc;
	HorizonInvRot(bam_vector,bam_vct_loc);
	reqpx = -atan(bam_vct_loc.z/bam_vct_loc.y);
	reqpz = atan(aspdv.x/aspdv.y);
	}
	else if (alt > 230)
	{
	VECTOR3 crus_vector = _V(0, -1, 0);
	VECTOR3 crus_vct_loc;
	HorizonInvRot(crus_vector,crus_vct_loc);
	reqpx = -atan(crus_vct_loc.z/crus_vct_loc.y);
	reqpz = atan(crus_vct_loc.x/crus_vct_loc.y);
	}

	double timex = (2*reqpx/ avel.x);
	double accx = -avel.x/timex;

	double timez = (2*reqpz/ avel.z);
	double accz = -avel.z/timez;

	double maxaccx = torquex/inrtx;
	double maxaccz = torquez/inrtz;


	double a;
	double b;
	double f = 10;

	int doff = 0;

	if (pitch > PI/2){
		doff = 1;
		ActivateNavmode(2);}
	if (pitch < -PI/2){
		doff = 1;
		ActivateNavmode(2);}
	if (bank > PI/2){
		doff = 1;
		ActivateNavmode(2);}
	if (bank < -PI/2){
		doff = 1;
		ActivateNavmode(2);}

	// do x axis

	if (doff == 0)
	{

	if (avel.x < 0) avel.x = -avel.x;

	if (timex < 0) a = (reqpx/f*PI);

	else if (timex > avel.x*10/maxaccx) a = (reqpx/f*PI);

	else a = accx/maxaccx;

	// do z axis

	if (avel.z < 0) avel.z = -avel.z;

	if (timez < 0) b = (reqpz/f*PI);

	else if (timez > avel.z*10/maxaccz) b = (reqpz/f*PI);

	else  b = accz/maxaccz;


	//

	double time = (-2*(alt-9)/(haspdv.y+1.5));
	double acc = ((-haspdv.y+1.5)/time);

	if (alt > 400)
	{
		time = (-2*(alt-9)/(haspdv.y+30));
		acc = ((-haspdv.y+30)/time);
	}

	double force = ((mass*acc)+ weight);
	double englvl = min(1, bf*pf*force/3633);

	a = a/2;
	b = b/2;

	double lvl1 = max(0,(englvl + a + a/3));
	double lvl2 =  max(0,(englvl - a + b + a/3));
	double lvl3 =  max(0,(englvl - a - b + a/3));

	lvl1 = min(1,lvl1);
	lvl2 =  min(1,lvl2);
	lvl3 =  min(1,lvl3);

	SetThrusterLevel(th_rcs[4], lvl1);
	SetThrusterLevel(th_rcs[5], lvl2);
	SetThrusterLevel(th_rcs[6], lvl3);
	}

	else
	{
	double time = (-2*(alt-9)/(haspdv.y+1.5));
	double acc = ((-haspdv.y+1.5)/time);
	double force = ((mass*acc)+ weight);
	double englvl = min(1, bf*pf*force/3633);
	englvl = max(0,englvl);
	SetThrusterLevel(th_rcs[4], englvl);
	SetThrusterLevel(th_rcs[5], englvl);
	SetThrusterLevel(th_rcs[6], englvl);
	}
}



void mpxl::DoOrient (void)
{
	double hdg;
	oapiGetHeading(GetHandle(),&hdg);
	hdg = hdg - PI;
	//if (hdg > PI) hdg = hdg -(2*PI);

	double mass = GetMass();
	VECTOR3 avel;
	GetAngularVel(avel);
	double inrt = (0.4*mass);
	double am = (0.4*mass*avel.y);
	double yawlf = (3.897);
	double yawrf = (4.554);

	double time = (2*hdg/ avel.y);
	double acc = avel.y/time;
	double mom = inrt*acc;
	double maxaccl = yawlf/inrt; //+ rotsp
	double maxaccr = yawrf/inrt; //- rotsp


	if (hdg > 0.01 && hdg < PI)
	{
		//DeactivateNavmode(1);
		//double thrust = min(1,(mom));
		double thrust = min(1,(mom/4.554));

		if (time < 0)
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,1);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}
		else if (time > avel.y/maxaccr)
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,1);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}
		else
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,thrust);
		}
	}
	else if (hdg > -PI && hdg < -0.01)
	{
		//DeactivateNavmode(1);
		//double thrust = min(1,(-mom));
		double thrust = min(1,(-mom/3.897));

		if (time < 0)
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,1);
		}
		else if (time > avel.y/-maxaccl)
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,1);
		}
		else
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,thrust);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}

	}
	else
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		ActivateNavmode(1);
		if (avel.y < 0.001 && avel.y > -0.001) DeactivateNavmode(1);
		else ActivateNavmode(1);
	}
}

void mpxl::DoDescent (void)
{
VECTOR3 weightv;
double weight;
GetWeightVector(weightv);
weight = length(weightv);

VECTOR3 haspdv;
GetGroundspeedVector(FRAME_HORIZON,haspdv);

VECTOR3 aspdv;
GetGroundspeedVector(FRAME_LOCAL,aspdv);

double pitch = GetPitch();
double cosp = (cos(pitch));
double sinp = (sin(pitch));
double pf = (1/(cos(pitch)));

double bank = GetBank();
double cosb = (cos(bank));
double bf = (1/(cos(bank)));

double ythr = GetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT);
double ythl = GetThrusterGroupLevel(THGROUP_ATT_YAWLEFT);
double ythf = ((ythr*7.794) + (ythl*7.794));


double alt = GetAltitude(ALTMODE_GROUND);

double mass = GetMass();

double hrz_vel = length(_V(haspdv.x, 0, haspdv.z));

if (alt < 0.8)
{
	SetThrusterGroupLevel(THGROUP_MAIN, 0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	DeactivateNavmode(2);
	DeactivateNavmode(1);
}
else if (haspdv.y < 0)
{

	if (9 > alt)
	{
	double acc = (-2*(haspdv.y + 1.5));
	double force = ((mass*acc)+ weight + ythf);
	double englvl = min(1, force/3633);
	englvl = max(0,englvl);
	SetThrusterGroupLevel(THGROUP_MAIN, englvl);

	DoOrient();
	ActivateNavmode(2);
	DeactivateNavmode(1);
	}

	else if (730 > alt)
	{
		DeactivateNavmode(2);
		//if (pitch < PI/4 && pitch > -PI/4 && bank < PI/4 && bank > -PI/4)DeactivateNavmode(2);
		DoAttitude();
		DoOrient();
	}
	else if (780 > alt) ActivateNavmode(2);
	/*else if (250 > alt)
	{
		if (pitch < PI/4 && pitch > -PI/4 && bank < PI/4 && bank > -PI/4)DeactivateNavmode(2);
		DoAttitude();
		DoOrient();
	}
	else if (720 > alt)
	{

	if (hrz_vel < 20) ActivateNavmode(1);
	else ActivateNavmode(2);

	double time = (-2*(alt-9)/(haspdv.y+1.5));

	double acc = ((-haspdv.y+2)/time);

	double force = ((mass*acc)+ weight);

	double englvl = min(1, pf*force/3633);

	englvl = max(0,englvl);

	SetThrusterGroupLevel(THGROUP_MAIN, englvl);
	}
	else if (770 > alt)
	{
	DeactivateNavmode(2);
	if ( thsw2 == 0) 
		{
		SetThrusterLevel(th_rcs[4], 0.1);
		SetThrusterLevel(th_rcs[5], 0);
		SetThrusterLevel(th_rcs[6], 0);
	thsw2 = 1;
		}
	}
	else if (800 > alt)ActivateNavmode(2);*/
}
else SetThrusterGroupLevel(THGROUP_MAIN, 0.0);

}

void mpxl::AddSmoke (void)
{
SURFHANDLE dust = oapiRegisterExhaustTexture ("Phoenix_Mars_Lander/mpxl_dust");

PARTICLESTREAMSPEC crashps = {
		0, 0.5, 15, 10, 1, 1, 2, 0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-2, 1, dust
	};

		AddExhaustStream(th_rcs[4],_V(0,-4,0.645), &crashps);
		AddExhaustStream(th_rcs[5],_V(-0.5586,-4,-0.3225), &crashps);
		AddExhaustStream(th_rcs[6],_V(0.5586,-4,-0.3225), &crashps);

}

MATRIX3 mpxl::RotationMatrix(VECTOR3 angles, bool xyz)
{
	MATRIX3 m;
	MATRIX3 RM_X, RM_Y, RM_Z;
	RM_X = _M(1, 0, 0, 0, cos(angles.x), -sin(angles.x), 0, sin(angles.x), cos(angles.x));
	RM_Y = _M(cos(angles.y), 0, sin(angles.y), 0, 1, 0, -sin(angles.y), 0, cos(angles.y));
	RM_Z = _M(cos(angles.z), -sin(angles.z), 0, sin(angles.z), cos(angles.z), 0, 0, 0, 1);
	if (!xyz) 
	{
		m = mul(RM_Z, mul(RM_Y, RM_X));
	}
	else 
	{
		m = mul(RM_X, mul(RM_Y, RM_Z));
	}
	return m;
}

void mpxl::SpawnRocks (void)
{
VESSELSTATUS2 vs2;
memset(&vs2, 0, sizeof(vs2));
vs2.version = 2;
GetStatusEx(&vs2);

double lng, lat, rad, hdg;
GetEquPos(lng,lat,rad);
oapiGetHeading(GetHandle(), &hdg);
vs2.status = 1;
vs2.surf_lng = lng;
vs2.surf_lat = lat;
vs2.surf_hdg = hdg;

//VECTOR3 surf_norm = GetSurfaceNormal();
//double lng_rot = atan2(surf_norm.x, surf_norm.y);
//double lat_rot = atan2(surf_norm.z, surf_norm.y);
//lng = lng + lng_rot;
//lat = lat + lat_rot;

MATRIX3 rot1 = RotationMatrix(_V(0 * RAD, (90 * RAD - lng), 0 * RAD), TRUE);
MATRIX3 rot2 = RotationMatrix(_V(-lat + 0 * RAD, 0, 0 * RAD), TRUE);
MATRIX3 rot3 = RotationMatrix(_V(0, 0, 180 * RAD + hdg), TRUE);
MATRIX3 rot4 = RotationMatrix(_V(90*RAD, 0, 0), TRUE);

MATRIX3 RotMatrix_Def = mul(rot1, mul(rot2, mul(rot3, rot4)));

vs2.arot.x = atan2(RotMatrix_Def.m23, RotMatrix_Def.m33);
vs2.arot.y = -asin(RotMatrix_Def.m13);
vs2.arot.z = atan2(RotMatrix_Def.m12, RotMatrix_Def.m11);
vs2.vrot.x = 0.05 ;
vs2.vrot.y = 0;
vs2.vrot.z = 0;

oapiCreateVesselEx("Rocks", "Phoenix_Mars_Lander/phoenix_rocks", &vs2);
rock_sw = 1;
}

void mpxl::clbkPreStep(double SimT, double SimDT, double MJD)
{
	// next block of code

	if (5 > GetAltitude(ALTMODE_GROUND) && smoke_sw == 0) 
	{
			AddSmoke();
			smoke_sw = 1;
	}
	//if (11000 > GetAltitude() && legs_status == 0) RevertLegs();
	if (legsw == 0){
	legs_proc = 1.0;
	legs_status = DOOR_OPEN;
	legsw = 1;
	SetAnimation (anim_legs, legs_proc);
	}

	if (500 > GetAltitude(ALTMODE_GROUND) && rock_sw == 0)SpawnRocks();

	if (740 > GetAltitude(ALTMODE_GROUND) && manov == 0) DoDescent();

	if (crashsw == 1)
	{
	if (oapiGetSimTime() > (crtime + 5)) oapiDelAnnotation (hNote);
	}

	if (GetAltitude(ALTMODE_GROUND) < 1 && GetAirspeed() > 10) DoCrash();

	else if (GroundContact() && GetAirspeed() > 10) DoCrash();

	//feet
	if (GroundContact() && feet_status == 0)RevertFeet();

	SetPropellantMass(ph_lid, 1);
	if (lidsw == 1) {

		if (strb == 0) SetThrusterLevel(th_lid, 1);
		else SetThrusterLevel(th_lid, 0);
		if (strb == 0) strb = 1;
		else if (strb == 1) strb = 0;
	}
	else SetThrusterLevel(th_lid, 0);

	////// animations //////////////

		// Animate solar arrays
	if (ant_status >= DOOR_CLOSING) {
		double da = SimDT * ANTENNA_OPERATING_SPEED;
		if (ant_status == DOOR_CLOSING) {
			if (ant_proc > 0.0) ant_proc = max (0.0, ant_proc-da);
			else                ant_status = DOOR_CLOSED;
		} else {
			if (ant_proc < 1.0) ant_proc = min (1.0, ant_proc+da);
			else                ant_status = DOOR_OPEN;
		}
		SetAnimation (anim_ant, ant_proc);
	}

	// Animate camera
	//if (fold_status >= DOOR_CLOSING) {
	//	double da = simdt * ARRAY_OPERATING_SPEED;
	//	if (fold_status == DOOR_CLOSING) {
	//		if (fold_proc > 0.0) fold_proc = max (0.0, fold_proc-da);
	//		else                 fold_status = DOOR_CLOSED;
	//	} else {
	//		if (fold_proc < 1.0) fold_proc = min (1.0, fold_proc+da);
	//		else                 fold_status = DOOR_OPEN;
	//	}
	//	SetAnimation (anim_fold, fold_proc);
	//}

	// animate legs
	if (legs_status >= DOOR_CLOSING) {
		double da = SimDT * LEGS_OPERATING_SPEED;
		if (legs_status == DOOR_CLOSING) {
			if (legs_proc > 0.0) legs_proc = max (0.0, legs_proc-da);
			else                 legs_status = DOOR_CLOSED;
		} else {
			if (legs_proc < 1.0) legs_proc = min (1.0, legs_proc+da);
			else                 legs_status = DOOR_OPEN;
		}
		SetAnimation (anim_legs, legs_proc);
	}

	// animate cover
	if (cov_status >= DOOR_CLOSING) {
		double da = SimDT * COV_OPERATING_SPEED;
		if (cov_status == DOOR_CLOSING) {
			if (cov_proc > 0.0) cov_proc = max (0.0, cov_proc-da);
			else                 cov_status = DOOR_CLOSED;
		} else {
			if (cov_proc < 1.0) cov_proc = min (1.0, cov_proc+da);
			else                 cov_status = DOOR_OPEN;
		}
		SetAnimation (anim_cov, cov_proc);
	}

	// animate feet
	if (feet_status >= DOOR_CLOSING) {
		double da = SimDT * FEET_OPERATING_SPEED;
		if (feet_status == DOOR_CLOSING) {
			if (feet_proc > 0.0) feet_proc = max (0.0, feet_proc-da);
			else                 feet_status = DOOR_CLOSED;
		} else {
			if (feet_proc < 1.0) feet_proc = min (1.0, feet_proc+da);
			else                 feet_status = DOOR_OPEN;
		}
		SetAnimation (anim_feet, feet_proc);
	}

	//Animate camera

	if (camrot == 1)
	{
	double da = SimDT * ARRAY_OPERATING_SPEED;
	fold_proc = fold_proc + da;
	if (fold_proc > 1) fold_proc = fold_proc - 1;
	SetAnimation (anim_fold, fold_proc);
	}

	else fold_proc = fold_proc;

	// animate arm rot
	if (armrotsw == 1)
	{
	double da = SimDT * ARMROT_SPEED;
	arp_proc = arp_proc + da;
	if (arp_proc > 1) arp_proc = 1;
	if (arp_proc < 0) arp_proc = 0;
	SetAnimation (anim_arp, arp_proc);
	armrotsw = 0;
	}

	else if (armrotsw == 2)
	{
	double da = SimDT * ARMROT_SPEED;
	arp_proc = arp_proc - da;
	if (arp_proc < 0) arp_proc = 0;
	if (arp_proc > 1) arp_proc = 1;
	SetAnimation (anim_arp, arp_proc);
	armrotsw = 0;
	}

	else arp_proc = arp_proc;

	// animate sld rot
	if (armsldsw == 1)
	{
	double da = SimDT * ARMSLD_SPEED;
	sld_proc = sld_proc + da;
	if (sld_proc > 1) sld_proc = 1;
	if (sld_proc < 0) sld_proc = 0;
	SetAnimation (anim_sld, sld_proc);
	armsldsw = 0;
	}

	else if (armsldsw == 2)
	{
	double da = SimDT * ARMROT_SPEED;
	sld_proc = sld_proc - da;
	if (sld_proc < 0) sld_proc = 0;
	if (sld_proc > 1) sld_proc = 1;
	SetAnimation (anim_sld, sld_proc);
	armsldsw = 0;
	}

	else sld_proc = sld_proc;

	// animate elbow rot
	if (armelbsw == 1)
	{
	double da = SimDT * ARMSLD_SPEED;
	elb_proc = elb_proc + da;
	if (elb_proc > 1) elb_proc = 1;
	if (elb_proc < 0) elb_proc = 0;
	SetAnimation (anim_elb, elb_proc);
	armelbsw = 0;
	}

	else if (armelbsw == 2)
	{
	double da = SimDT * ARMSLD_SPEED;
	elb_proc = elb_proc - da;
	if (elb_proc < 0) elb_proc = 0;
	if (elb_proc > 1) elb_proc = 1;
	SetAnimation (anim_elb, elb_proc);
	armelbsw = 0;
	}

	else elb_proc = elb_proc;

}


bool mpxl::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);


	
	{
	if (manov == 0)skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x006666FF);
	char kbuf[256];
	sprintf(kbuf,"M = Descent Manual Override");
	skp->Text ((10+roxl), (wc+royl), kbuf,strlen(kbuf));
	}
	

	if (cov_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"G = Deploy MET Antenna, Arm, Camera, Lidar");
	skp->Text ((10+roxl), (w0+royl), kbuf,strlen(kbuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"C = Rotate Camera");
	skp->Text ((10+roxl), (w0+royl), kbuf,strlen(kbuf));

	char bbuf[256];
	sprintf(bbuf,"D = Lidar On/Off");
	skp->Text ((10+roxl), (w1+royl), bbuf,strlen(bbuf));

	char cbuf[256];
	sprintf(cbuf,"B / Shift+B = Deploy / Retract Arm");
	skp->Text ((10+roxl), (w2+royl), cbuf,strlen(cbuf));

	char dbuf[256];
	sprintf(dbuf,"V / Shift+V = Rotate Arm");
	skp->Text ((10+roxl), (w3+royl), dbuf,strlen(dbuf));

	char ebuf[256];
	sprintf(ebuf,"N / Shift+N = Arm Elbow");
	skp->Text ((10+roxl), (w4+royl), ebuf,strlen(ebuf));
	}

	if (ant_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"K = Deploy Solar Panels");
	skp->Text ((10+roxl), (wd+royl), kbuf,strlen(kbuf));
	}

	return true;

}


void mpxl::clbkPostStep (double simt, double simdt, double mjd)
{

	

}